from copy import deepcopy
import random
from typing import Optional, Tuple

import numpy as np
import torch
import torch.nn.functional as F
import torch.nn as nn
from tqdm import trange

from dataset_helpers import get_dataloaders, get_stiffness_batches
from experiment_config import (
    Config,
    DatasetSubsetType,
    HParams,
    State,
    EvaluationMetrics,
    OptimizerType,
)
from logs import BaseLogger, Printer
from measures import get_all_measures
from stiffness import stiffness, gradient_variance
from models import NiN, CustomResNet, p_resnet_initialiser, skipinit_resnet_initialiser, mlp_initialiser, \
    cnn_initialiser
from custom_optimizers import SGLD

STIFFNESS_SET_SIZE = 500


def calculate_gradient_norm(model):
    grad_norm = 0
    for param in model.parameters():
        grad_norm += param.grad.pow(2).sum()
    return grad_norm


class Experiment:
    def __init__(
            self,
            state: State,
            device: torch.device,
            hparams: HParams,
            config: Config,
            logger: BaseLogger,
            result_save_callback: Optional[object] = None
    ):
        self.state = state
        self.device = device
        self.hparams = hparams
        self.config = config

        # Random Seeds
        random.seed(self.hparams.seed)
        np.random.seed(self.hparams.seed)
        torch.manual_seed(self.hparams.seed)
        torch.cuda.manual_seed_all(self.hparams.seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

        # Logging
        self.logger = logger
        self.train_original_losses = list()
        self.train_modified_losses = list()
        # Printing
        self.printer = Printer(self.config.id, self.config.verbosity)
        self.result_save_callback = result_save_callback

        # Model
        if self.hparams.model_arch == "nin":
            self.model = NiN(self.hparams.model_depth, self.hparams.model_width, self.hparams.base_width,
                         self.hparams.dataset_type)
            print("Using network in network architecture")
        elif self.hparams.model_arch == "resnet":
            self.model = CustomResNet(self.hparams.model_depth, self.hparams.model_width,
                                      self.hparams.base_width, self.hparams.dataset_type)
            print("Using resnet architecture meant for ImageNet")
        elif self.hparams.model_arch == "p_resnet":
            self.model = p_resnet_initialiser(self.hparams.model_depth, self.hparams.model_width,
                                              self.hparams.base_width, self.hparams.dataset_type)
            print("Using resnet architecture meant for CIFAR")
        elif self.hparams.model_arch == "skiptinit_resnet":
            self.model = skipinit_resnet_initialiser(self.hparams.model_depth, self.hparams.model_width,
                                                     self.hparams.dataset_type)
            print("Using skipinit resnet architecture meant for CIFAR")
        elif self.hparams.model_arch == "mlp":
            self.model = mlp_initialiser(self.hparams.model_depth, self.hparams.model_width, self.hparams.dataset_type)
        elif self.hparams.model_arch == "cnn":
            self.model = cnn_initialiser(self.hparams.model_depth, self.hparams.model_width, self.hparams.dataset_type)
        else:
            print(self.hparams.model_arch)
            raise AssertionError

        print("Number of parameters", sum(p.numel() for p in self.model.parameters() if p.requires_grad))
        self.model.to(device)
        self.init_model = deepcopy(self.model)

        # Optimizer
        if self.hparams.optimizer_type == OptimizerType.SGD:
            self.optimizer = torch.optim.SGD(self.model.parameters(), lr=self.hparams.lr)
        elif self.hparams.optimizer_type == OptimizerType.SGD_MOMENTUM:
            self.optimizer = torch.optim.SGD(self.model.parameters(), lr=self.hparams.lr, momentum=0.9)
        elif self.hparams.optimizer_type == OptimizerType.SGD_WEIGHT_DECAY:
            self.optimizer = torch.optim.SGD(self.model.parameters(), lr=self.hparams.lr,
                                             weight_decay=self.hparams.wd)
        elif self.hparams.optimizer_type == OptimizerType.SGD_MOMENTUM_WEIGHT_DECAY:
            self.optimizer = torch.optim.SGD(self.model.parameters(), lr=self.hparams.lr, momentum=0.9,
                                             weight_decay=self.hparams.wd)
        elif self.hparams.optimizer_type == OptimizerType.ADAM:
            self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.hparams.lr)
        elif self.hparams.optimizer_type == OptimizerType.SGLD:
            self.optimizer = SGLD(self.model.parameters(), lr=self.hparams.lr,
                                  dataset_size=self.hparams.train_dataset_size)
        else:
            raise KeyError

        # LR scheduling
        if self.hparams.lr_scheduling:
            self.lr_scheduler = torch.optim.lr_scheduler.MultiStepLR(
                self.optimizer, milestones=[100, 150], last_epoch=self.state.epoch - 2)

        # Load data
        self.train_loader, self.train_eval_loader, self.test_loader = get_dataloaders(self.hparams, self.config,
                                                                                      self.device)
        # Load stiffness testing pairs
        (self.train_stiffness_data, self.train_stiffness_target), \
            (self.train_stiffness_data2, self.train_stiffness_target2), \
            (self.valid_stiffness_data, self.valid_stiffness_target) = \
            get_stiffness_batches(self.train_eval_loader, self.test_loader, STIFFNESS_SET_SIZE)
        self.stiffness_criterion = nn.CrossEntropyLoss(reduction='none')

    def save_state(self, postfix: str = '') -> None:
        checkpoint_file = self.config.checkpoint_dir / (self.hparams.md5 + postfix + '.pt')
        torch.save({
            'config': self.hparams,
            'state': self.state,
            'model': self.model.state_dict(),
            'init_model': self.init_model.state_dict(),
            'optimizer': self.optimizer.state_dict(),
            'np_rng': np.random.get_state(),
            'torch_rng': torch.get_rng_state(),
        }, checkpoint_file)

    def _train_epoch(self) -> Tuple[float, float]:
        self.model.train()
        ce_check_batches = [(len(self.train_loader) // (2 ** (self.state.ce_check_freq))) * (i + 1) for i in
                            range(2 ** (self.state.ce_check_freq) - 1)]
        ce_check_batches.append(len(self.train_loader) - 1)

        epoch_sum_original_loss = 0.
        epoch_sum_modified_loss = 0.
        epoch_batches = 0
        for batch_idx, (data, target) in enumerate(self.train_loader):
            epoch_batches += 1
            data, target = data.to(self.device, non_blocking=True), target.to(self.device, non_blocking=True)

            self.state.batch = batch_idx
            self.state.global_batch = (self.state.epoch - 1) * len(self.train_loader) + self.state.batch
            self.model.train()

            self.optimizer.zero_grad()

            logits = self.model(data)
            if self.hparams.K_train > 0:
                logits = torch.mean(logits.view(-1, self.hparams.K_train, 10), dim=1)
            else:
                pass

            original_loss = F.cross_entropy(logits, target)

            if self.hparams.gradient_norm_reg:
                original_loss.backward(create_graph=True)
                gradient_norm_regularisation = self.hparams.gradient_norm_penalty * calculate_gradient_norm(self.model)
                gradient_norm_regularisation.backward()
                modified_loss = original_loss + gradient_norm_regularisation
            else:
                original_loss.backward()
                modified_loss = original_loss
                gradient_norm_regularisation = None

            self.optimizer.step()

            # Log everything
            self.printer.batch_end(self.config, self.state, data, self.train_loader, modified_loss)
            self.logger.log_batch_end(self.config, self.state, original_loss, modified_loss)

            epoch_sum_original_loss += original_loss.cpu().item()
            epoch_sum_modified_loss += modified_loss.cpu().item()
            # print(original_loss.cpu().item())

            del original_loss, gradient_norm_regularisation, modified_loss

            # Cross-entropy stopping check
            if batch_idx == ce_check_batches[0]:
                ce_check_batches.pop(0)
                is_last_batch = batch_idx == (len(self.train_loader) - 1)
                dataset_ce = self.evaluate_cross_entropy(DatasetSubsetType.TRAIN, log=is_last_batch)[0]
                if dataset_ce < self.hparams.ce_target:
                    self.state.converged = True
                else:
                    while len(self.state.ce_check_milestones) > 0 and dataset_ce <= self.state.ce_check_milestones[0]:
                        passed_milestone = self.state.ce_check_milestones[0]
                        print(f'passed ce milestone {passed_milestone}')
                        self.state.ce_check_milestones.pop(0)
                        self.state.ce_check_freq += 1
                        if self.config.save_epoch_freq is not None:
                            self.save_state(f'_ce_{passed_milestone}')

            if self.state.converged:
                break
        return epoch_sum_original_loss / epoch_batches, epoch_sum_modified_loss / epoch_batches

    def train(self) -> None:
        self.printer.train_start(self.device)
        train_eval, val_eval = None, None

        if self.hparams.K_train > 0:
            total_epochs = int(self.hparams.epochs / self.hparams.K_train)
        else:
            total_epochs = self.hparams.epochs
        self.state.global_batch = 0
        for epoch in trange(self.state.epoch, total_epochs + 1, disable=(not self.config.use_tqdm)):
            self.state.epoch = epoch
            epoch_train_original_loss, epoch_train_modified_loss = self._train_epoch()
            # LR scheduling
            if self.hparams.lr_scheduling:
                self.lr_scheduler.step()
            # Log the average training loss for the epoch
            self.logger.log_metrics(
                self.state.global_batch,
                {'epoch_average_training_original_loss': epoch_train_original_loss,
                 'epoch_average_training_modified_loss': epoch_train_modified_loss}
            )
            self.train_original_losses.append(epoch_train_original_loss)
            self.train_modified_losses.append(epoch_train_modified_loss)

            is_evaluation_epoch = (
                        epoch == 1 or epoch == total_epochs or epoch % self.config.log_epoch_freq == 0)
            is_stiffness_evaluation_epoch = (
                    epoch == 1 or epoch == 5 or self.state.converged or epoch == total_epochs)
            if is_evaluation_epoch or self.state.converged:
                train_eval = self.evaluate(DatasetSubsetType.TRAIN,
                                           (epoch == total_epochs or self.state.converged))
                val_eval = self.evaluate(DatasetSubsetType.TEST)
                self.logger.log_generalization_gap(self.state, train_eval.acc, val_eval.acc, train_eval.avg_loss,
                                                   val_eval.avg_loss, train_eval.all_complexities)
                self.printer.epoch_metrics(epoch, train_eval, val_eval)

            if is_stiffness_evaluation_epoch:
                train_train_stiffness_sign_expectation, train_train_stiffness_cos_expectation = \
                    stiffness(self.model,
                              self.train_stiffness_data, self.train_stiffness_target,
                              self.train_stiffness_data2, self.train_stiffness_target2,
                              self.stiffness_criterion, self.device)

                train_val_stiffness_sign_expectation, train_val_stiffness_cos_expectation = \
                    stiffness(self.model,
                              self.train_stiffness_data, self.train_stiffness_target,
                              self.valid_stiffness_data, self.valid_stiffness_target,
                              self.stiffness_criterion, self.device)
                self.logger.log_metrics(
                    self.state.global_batch,
                    {'train_train_stiffness_sign': train_train_stiffness_sign_expectation,
                     'train_train_stiffness_cos': train_train_stiffness_cos_expectation,
                     'train_val_stiffness_sign': train_val_stiffness_sign_expectation,
                     'train_val_stiffness_cos': train_val_stiffness_cos_expectation}
                )
                gradient_var = gradient_variance(model=self.model, dataloader=self.train_eval_loader,
                                                 criterion=self.stiffness_criterion, device=self.device)
                self.logger.log_metrics(
                    self.state.global_batch,
                    {
                        "gradient_variance": gradient_var if isinstance(gradient_var, float) else gradient_var.cpu().item()
                    }
                )

            if epoch == total_epochs or self.state.converged:
                self.result_save_callback(epoch, val_eval, train_eval)

            # Save state
            is_save_epoch = self.config.save_epoch_freq is not None and (
                        epoch % self.config.save_epoch_freq == 0 or epoch == total_epochs or self.state.converged)
            if is_save_epoch:
                self.save_state()

            if self.state.converged:
                print('Converged')
                break

        self.printer.train_end()

        if train_eval is None or val_eval is None:
            raise RuntimeError

    def evaluate(self, dataset_subset_type: DatasetSubsetType, compute_all_measures: bool = False) -> EvaluationMetrics:
        self.model.eval()
        data_loader = [self.train_eval_loader, self.test_loader][dataset_subset_type]

        cross_entropy_loss, acc, num_correct = self.evaluate_cross_entropy(dataset_subset_type)

        all_complexities = {}
        if dataset_subset_type == DatasetSubsetType.TRAIN and compute_all_measures:
            all_complexities = get_all_measures(self.model, self.init_model, data_loader, acc, self.hparams.seed,
                                                self.train_original_losses, self.hparams)

        self.logger.log_epoch_end(self.hparams, self.state, dataset_subset_type, cross_entropy_loss, acc)

        return EvaluationMetrics(acc, cross_entropy_loss, num_correct, len(data_loader.dataset), all_complexities)

    @torch.no_grad()
    def evaluate_cross_entropy(self, dataset_subset_type: DatasetSubsetType, log: bool = False):
        self.model.eval()
        cross_entropy_loss = 0
        num_correct = 0

        data_loader = [self.train_eval_loader, self.test_loader][dataset_subset_type]
        num_to_evaluate_on = len(data_loader.dataset)

        for data, target in data_loader:
            data, target = data.to(self.device, non_blocking=True), target.to(self.device, non_blocking=True)
            logits = self.model(data)
            if self.hparams.K_train > 0:
                logits = torch.mean(logits.view(-1, self.hparams.K_train, 10), dim=1)
            else:
                pass
            cross_entropy = F.cross_entropy(logits, target, reduction='sum')
            cross_entropy_loss += cross_entropy.item()  # sum up batch loss

            pred = logits.data.max(1, keepdim=True)[1]  # get the index of the max logits
            batch_correct = pred.eq(target.data.view_as(pred)).type(torch.FloatTensor).cpu()
            num_correct += batch_correct.sum()

        cross_entropy_loss /= num_to_evaluate_on
        acc = num_correct.item() / num_to_evaluate_on

        if log:
            self.logger.log_epoch_end(self.hparams, self.state, dataset_subset_type, cross_entropy_loss, acc)

        self.logger.log_metrics(
            self.state.global_batch,
            {'{}_eval_loss'.format(["TRIAN", "TEST"][dataset_subset_type]): cross_entropy_loss,
             '{}_eval_acc'.format(["TRIAN", "TEST"][dataset_subset_type]): acc}
        )
        return cross_entropy_loss, acc, num_correct
