from dataclasses import asdict, dataclass, field
from enum import Enum, IntEnum
import hashlib
from pathlib import Path
import time
from typing import Dict, List, NamedTuple, Optional, Tuple


class DatasetType(Enum):
  CIFAR10 = (1, (3, 32, 32), 10)
  SVHN = (2, (3, 32, 32), 10)
  
  def __init__(self, id: int, image_shape: Tuple[int, int, int], num_classes: int):
    self.D = image_shape
    self.K = num_classes

class DatasetSubsetType(IntEnum):
  TRAIN = 0
  TEST = 1

class ComplexityType(Enum):
  # Measures from Fantastic Generalization Measures (equation numbers)
  PARAMS = 20
  INVERSE_MARGIN = 22
  LOG_SPEC_INIT_MAIN = 29
  LOG_SPEC_ORIG_MAIN = 30
  LOG_PROD_OF_SPEC_OVER_MARGIN = 31
  LOG_PROD_OF_SPEC = 32
  FRO_OVER_SPEC = 33
  LOG_SUM_OF_SPEC_OVER_MARGIN = 34
  LOG_SUM_OF_SPEC = 35
  LOG_PROD_OF_FRO_OVER_MARGIN = 36
  LOG_PROD_OF_FRO = 37
  LOG_SUM_OF_FRO_OVER_MARGIN = 38
  LOG_SUM_OF_FRO = 39
  FRO_DIST = 40
  DIST_SPEC_INIT = 41
  PARAM_NORM = 42
  PATH_NORM_OVER_MARGIN = 43
  PATH_NORM = 44
  PACBAYES_INIT = 48
  PACBAYES_ORIG = 49
  PACBAYES_FLATNESS = 53
  PACBAYES_MAG_INIT = 56
  PACBAYES_MAG_ORIG = 57
  PACBAYES_MAG_FLATNESS = 61
  # Other Measures
  L2 = 100
  L2_DIST = 101
  # FFT Spectral Measures
  LOG_SPEC_INIT_MAIN_FFT = 129
  LOG_SPEC_ORIG_MAIN_FFT = 130
  LOG_PROD_OF_SPEC_OVER_MARGIN_FFT = 131
  LOG_PROD_OF_SPEC_FFT = 132
  FRO_OVER_SPEC_FFT = 133
  LOG_SUM_OF_SPEC_OVER_MARGIN_FFT = 134
  LOG_SUM_OF_SPEC_FFT = 135
  DIST_SPEC_INIT_FFT = 141
  # Training Speed Measures
  SUM_OVER_TRAINING_LOSSES = 200
  SUM_OVER_TRAINING_LOSSES_E_1 = 201
  SUM_OVER_TRAINING_LOSSES_E_10 = 210
  SUM_OVER_TRAINING_LOSSES_E_20 = 220
  SUM_OVER_TRAINING_LOSSES_E_30 = 230
  SUM_OVER_TRAINING_LOSSES_E_40 = 240
  SUM_OVER_TRAINING_LOSSES_E_50 = 250
  SUM_OVER_TRAINING_LOSSES_E_60 = 260
  SUM_OVER_TRAINING_LOSSES_E_70 = 270
  # Hessian eigenvalue flatness mesures
  TOP_EIGENVALUE_1 = 281
  TOP_EIGENVALUE_2 = 282
  TOP_EIGENVALUE_3 = 283
  TOP_EIGENVALUE_4 = 284
  TOP_EIGENVALUE_5 = 285
  # Value sensitivity, gradient variance, and gradient sensitivity
  VALUE_SENSITIVITY = 300
  GRADIENT_VARIANCE = 301
  GRADIENT_SENSITIVITY = 302

class OptimizerType(Enum):
  SGD = 1
  SGD_MOMENTUM = 2
  SGD_MOMENTUM_WEIGHT_DECAY = 3
  ADAM = 4
  SGD_WEIGHT_DECAY = 5
  SGLD = 6

class Verbosity(IntEnum):
  NONE = 1
  RUN = 2
  EPOCH = 3
  BATCH = 4

@dataclass(frozen=False)
class State:
  epoch: int = 1
  batch: int = 1
  global_batch: int = 1
  converged: bool = False
  ce_check_freq: int = 0
  ce_check_milestones: Optional[List[float]] = None

# Hyperparameters that uniquely determine the experiment
@dataclass(frozen=True)
class HParams:
  seed: int = 0
  use_cuda: bool = True
  # Model
  model_depth: int = 2
  model_width: int = 8
  base_width: int = 25
  # Dataset
  dataset_type: DatasetType = DatasetType.CIFAR10
  data_seed: Optional[int] = 42
  train_dataset_size: Optional[int] = None
  test_dataset_size: Optional[int] = None
  transforms: bool = False
  # Training
  batch_size: int = 32
  epochs: int = 300
  optimizer_type: OptimizerType = OptimizerType.SGD_MOMENTUM
  lr: float = 0.01
  wd: float = 1e-4
  lr_scheduling : bool = False
  gradient_norm_reg : bool = False
  # Cross-entropy stopping criterion
  ce_target: Optional[float] = 0.01
  ce_target_milestones: Optional[List[float]] = field(default_factory=lambda: [0.05, 0.025, 0.015])
  # Model architecture
  model_arch: str = "nin"
  # Gradient norm regularisation
  gradient_norm_penalty : float = 0
  # Augmentation
  K_train : int = -1

  def to_tensorboard_dict(self) -> dict:
    d = asdict(self)
    d = {x: y for (x,y) in d.items() if y is not None}
    d = {x:(y.name if isinstance(y, Enum) else y) for x,y in d.items()}
    return d
  
  @property
  def md5(self):
    return hashlib.md5(str(self).encode('utf-8')).hexdigest()
  
  @property
  def wandb_md5(self):
    dictionary = self.to_tensorboard_dict()
    dictionary['seed'] = 0
    return hashlib.md5(str(dictionary).encode('utf-8')).hexdigest()

# Configuration which doesn't affect experiment results
@dataclass(frozen=True)
class Config:
  id: int = field(default_factory=lambda: time.time_ns())
  log_batch_freq: Optional[int] = None
  log_epoch_freq: Optional[int] = 10
  save_epoch_freq: Optional[int] = 1
  root_dir: Path = Path('/scratch/ms20qj/large_models')
  data_dir: Path = Path('./temp/data')
  project_name: str = 'real_grad_reg'
  verbosity: Verbosity = Verbosity.EPOCH
  use_tqdm: bool = False

  def setup_dirs(self) -> None:
    # Set up directories
    for directory in ('results', 'checkpoints'):
      (self.root_dir / directory).mkdir(parents=True, exist_ok=True)
    self.data_dir.mkdir(parents=True, exist_ok=True)

  @property
  def checkpoint_dir(self):
    return self.root_dir / 'checkpoints'
  
  @property
  def results_dir(self):
    return self.root_dir / 'results'

class EvaluationMetrics(NamedTuple):
  acc: float
  avg_loss: float
  num_correct: int
  num_to_evaluate_on: int
  all_complexities: Dict[ComplexityType, float]
